from __future__ import annotations

import typing
from io import BytesIO
from pathlib import Path
from typing import Any, Type
from urllib.parse import urlparse

import boto3
import botocore.client  # pylint: disable=unused-import
from botocore.exceptions import ClientError

from digitalhub.readers.data.api import get_reader_by_object
from digitalhub.stores._base.store import Store
from digitalhub.stores.s3.configurator import S3StoreConfigurator
from digitalhub.stores.s3.utils import get_bucket_name
from digitalhub.utils.exceptions import StoreError
from digitalhub.utils.file_utils import get_file_info_from_s3, get_file_mime_type
from digitalhub.utils.types import SourcesOrListOfSources

if typing.TYPE_CHECKING:
    pass

# Type aliases
S3Client = Type["botocore.client.S3"]


class S3Store(Store):
    """
    S3 store class. It implements the Store interface and provides methods to fetch and persist
    artifacts on S3 based storage.
    """

    def __init__(self, config: dict | None = None) -> None:
        super().__init__()
        self._configurator = S3StoreConfigurator()
        self._configurator.configure(config)

    ##############################
    # I/O methods
    ##############################

    def download(
        self,
        root: str,
        dst: Path,
        src: list[str],
        overwrite: bool = False,
    ) -> str:
        """
        Download artifacts from storage.

        Parameters
        ----------
        root : str
            The root path of the artifact.
        dst : str
            The destination of the artifact on local filesystem.
        src : list[str]
            List of sources.
        overwrite : bool
            Specify if overwrite existing file(s).

        Returns
        -------
        str
            Destination path of the downloaded artifact.
        """
        client, bucket = self._check_factory(root)

        # Build destination directory
        if dst.suffix == "":
            dst.mkdir(parents=True, exist_ok=True)
        else:
            dst.parent.mkdir(parents=True, exist_ok=True)

        # Handle src and tree destination
        if self.is_partition(root):
            if not src:
                keys = self._list_objects(client, bucket, root)
                strip_root = self._get_key(root)
                trees = [k.removeprefix(strip_root) for k in keys]
            else:
                keys = self._build_key_from_root(root, src)
                trees = [s for s in src]
        else:
            keys = [self._get_key(root)]
            if not src:
                trees = [Path(self._get_key(root)).name]
            else:
                trees = [s for s in src]

        if len(keys) != len(trees):
            raise StoreError("Keys and trees must have the same length.")

        # Download files
        for elements in zip(keys, trees):
            key = elements[0]
            tree = elements[1]

            # Build destination path
            if dst.suffix == "":
                dst_pth = Path(dst, tree)
            else:
                dst_pth = dst

            # Check if destination path already exists
            self._check_overwrite(dst_pth, overwrite)

            self._build_path(dst_pth.parent)

            self._download_file(key, dst_pth, client, bucket)

        if len(trees) == 1:
            if dst.suffix == "":
                return str(Path(dst, trees[0]))
        return str(dst)

    def upload(
        self,
        src: SourcesOrListOfSources,
        dst: str,
    ) -> list[tuple[str, str]]:
        """
        Upload an artifact to storage.

        Parameters
        ----------
        src : SourcesOrListOfSources
            Source(s).
        dst : str
            The destination of the artifact on storage.

        Returns
        -------
        list[tuple[str, str]]
            Returns the list of destination and source paths of the uploaded artifacts.
        """
        # Destination handling
        key = self._get_key(dst)

        # Source handling (files list, dir or single file)
        src_is_list = isinstance(src, list)
        if not src_is_list:
            self._check_local_src(src)
            src_is_dir = Path(src).is_dir()
        else:
            for s in src:
                self._check_local_src(s)
            src_is_dir = False
            if len(src) == 1:
                src = src[0]

        # If source is a directory, destination must be a partition
        if (src_is_dir or src_is_list) and not dst.endswith("/"):
            raise StoreError(
                "If source is a list of files or a directory, "
                "destination must be a partition, e.g. 's3://bucket/partition/'"
            )

        # S3 client
        client, bucket = self._check_factory(dst)

        # Directory
        if src_is_dir:
            return self._upload_dir(src, key, client, bucket)

        # List of files
        elif src_is_list:
            return self._upload_file_list(src, key, client, bucket)

        # Single file
        return self._upload_single_file(src, key, client, bucket)

    def upload_fileobject(
        self,
        src: BytesIO,
        dst: str,
    ) -> str:
        """
        Upload an BytesIO to S3 based storage.

        Parameters
        ----------
        src : BytesIO
            The source object to be persisted.
        dst : str
            The destination path of the artifact.

        Returns
        -------
        str
            S3 key of the uploaded artifact.
        """
        client, bucket = self._check_factory(dst)
        key = self._get_key(dst)
        self._upload_fileobject(src, key, client, bucket)
        return f"s3://{bucket}/{key}"

    def get_file_info(
        self,
        root: str,
        paths: list[tuple[str, str]],
    ) -> list[dict]:
        """
        Method to get file metadata.

        Parameters
        ----------
        paths : list
            List of source paths.

        Returns
        -------
        list[dict]
            Returns files metadata.
        """
        client, bucket = self._check_factory(root)

        infos = []
        for i in paths:
            key, src_path = i

            # Rebuild key in case here arrive an s3://bucket prefix
            key = self._get_key(key)

            # Get metadata
            metadata = client.head_object(Bucket=bucket, Key=key)

            # Get file info
            info = get_file_info_from_s3(src_path, metadata)
            infos.append(info)

        return infos

    ##############################
    # Datastore methods
    ##############################

    def read_df(
        self,
        path: SourcesOrListOfSources,
        file_format: str | None = None,
        engine: str | None = None,
        **kwargs,
    ) -> Any:
        """
        Read DataFrame from path.

        Parameters
        ----------
        path : SourcesOrListOfSources
            Path(s) to read DataFrame from.
        file_format : str
            Extension of the file.
        engine : str
            Dataframe engine (pandas, polars, etc.).
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        Any
            DataFrame.
        """
        reader = self._get_reader(engine)

        # Verify if partition or single file
        if self.is_partition(path):
            client, bucket = self._check_factory(path)
            objects = self._list_objects(client, bucket, path)
            keys = [self._get_key(o) for o in objects]

        else:
            if isinstance(path, list):
                client, bucket = self._check_factory(path[0])
                keys = [self._get_key(p) for p in path]
            else:
                client, bucket = self._check_factory(path)
                keys = [self._get_key(path)]

        dfs = []
        for key in keys:
            file_format = self._get_extension(file_format, key)
            obj = self._download_fileobject(key, client, bucket)
            dfs.append(reader.read_df(obj, extension=file_format, **kwargs))

        if len(dfs) == 1:
            return dfs[0]
        return reader.concat_dfs(dfs)

    def query(
        self,
        query: str,
        engine: str | None = None,
    ) -> Any:
        """
        Query data from database.

        Parameters
        ----------
        query : str
            The query to execute.
        engine : str
            Dataframe engine (pandas, polars, etc.).

        Returns
        -------
        Any
            DataFrame.
        """
        raise StoreError("S3 store does not support query.")

    def write_df(
        self,
        df: Any,
        dst: str,
        extension: str | None = None,
        **kwargs,
    ) -> str:
        """
        Write a dataframe to S3 based storage. Kwargs are passed to df.to_parquet().

        Parameters
        ----------
        df : Any
            The dataframe.
        dst : str
            The destination path on S3 based storage.
        extension : str
            The extension of the file.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        str
            The S3 path where the dataframe was saved.
        """
        fileobj = BytesIO()
        reader = get_reader_by_object(df)
        reader.write_df(df, fileobj, extension=extension, **kwargs)
        return self.upload_fileobject(fileobj, dst)

    ##############################
    # Private I/O methods
    ##############################

    @staticmethod
    def _download_file(
        key: str,
        dst_pth: Path,
        client: S3Client,
        bucket: str,
    ) -> list[str]:
        """
        Download files from S3 partition.

        Parameters
        ----------
        key : str
            The key to be downloaded.
        dst_pth : str
            The destination of the files on local filesystem.
        client : S3Client
            The S3 client object.
        bucket : str
            The name of the S3 bucket.

        Returns
        -------
        list[str]
            The list of paths of the downloaded files.
        """
        client.download_file(bucket, key, dst_pth)

    @staticmethod
    def _download_fileobject(
        key: str,
        client: S3Client,
        bucket: str,
    ) -> BytesIO:
        """
        Download fileobject from S3 partition.

        Parameters
        ----------
        key : str
            The key of the file.
        client : S3Client
            The S3 client object.
        bucket : str
            The name of the S3 bucket.

        Returns
        -------
        BytesIO
            The fileobject of the downloaded file.
        """
        obj = client.get_object(Bucket=bucket, Key=key)
        return BytesIO(obj["Body"].read())

    def _upload_dir(
        self,
        src: str,
        dst: str,
        client: S3Client,
        bucket: str,
    ) -> list[tuple[str, str]]:
        """
        Upload directory to storage.

        Parameters
        ----------
        src : str
            List of sources.
        dst : str
            The destination of the artifact on storage.
        client : S3Client
            The S3 client object.
        bucket : str
            The name of the S3 bucket.

        Returns
        -------
        list[tuple[str, str]]
            Returns the list of destination and source paths of the uploaded artifacts.
        """
        # Get list of files
        src_pth = Path(src)
        files = [i for i in src_pth.rglob("*") if i.is_file()]

        # Build keys
        keys = []
        for i in files:
            i = i.relative_to(src_pth)
            keys.append(f"{dst}{i}")

        # Upload files
        paths = []
        for f, k in zip(files, keys):
            self._upload_file(f, k, client, bucket)
            paths.append((k, str(f.relative_to(src_pth))))
        return paths

    def _upload_file_list(
        self,
        src: list[str],
        dst: str,
        client: S3Client,
        bucket: str,
    ) -> list[tuple[str, str]]:
        """
        Upload list of files to storage.

        Parameters
        ----------
        src : list
            List of sources.
        dst : str
            The destination of the artifact on storage.
        client : S3Client
            The S3 client object.
        bucket : str
            The name of the S3 bucket.

        Returns
        -------
        list[tuple[str, str]]
            Returns the list of destination and source paths of the uploaded artifacts.
        """
        files = src
        keys = []
        for i in files:
            keys.append(f"{dst}{Path(i).name}")
        if len(set(keys)) != len(keys):
            raise StoreError("Keys must be unique (Select files with different names, otherwise upload a directory).")

        # Upload files
        paths = []
        for f, k in zip(files, keys):
            self._upload_file(f, k, client, bucket)
            paths.append((k, Path(f).name))
        return paths

    def _upload_single_file(
        self,
        src: str,
        dst: str,
        client: S3Client,
        bucket: str,
    ) -> str:
        """
        Upload a single file to storage.

        Parameters
        ----------
        src : str
            List of sources.
        dst : str
            The destination of the artifact on storage.
        client : S3Client
            The S3 client object.
        bucket : str
            The name of the S3 bucket.

        Returns
        -------
        str
            Returns the list of destination and source paths of the uploaded artifacts.
        """
        if dst.endswith("/"):
            dst = f"{dst.removeprefix('/')}{Path(src).name}"

        # Upload file
        self._upload_file(src, dst, client, bucket)
        name = Path(self._get_key(dst)).name
        return [(dst, name)]

    @staticmethod
    def _upload_file(
        src: str,
        key: str,
        client: S3Client,
        bucket: str,
    ) -> None:
        """
        Upload a file to S3 based storage. The function checks if the
        bucket is accessible.

        Parameters
        ----------
        src : str
            The source path of the file on local filesystem.
        key : str
            The key of the file on S3 based storage.
        client : S3Client
            The S3 client object.
        bucket : str
            The name of the S3 bucket.

        Returns
        -------
        None
        """
        extra_args = {}
        mime_type = get_file_mime_type(src)
        if mime_type is not None:
            extra_args["ContentType"] = mime_type
        client.upload_file(Filename=src, Bucket=bucket, Key=key, ExtraArgs=extra_args)

    @staticmethod
    def _upload_fileobject(
        fileobj: BytesIO,
        key: str,
        client: S3Client,
        bucket: str,
    ) -> None:
        """
        Upload a fileobject to S3 based storage. The function checks if the bucket is accessible.

        Parameters
        ----------
        fileobj : BytesIO
            The fileobject to be uploaded.
        key : str
            The key of the file on S3 based storage.
        client : S3Client
            The S3 client object.
        bucket : str
            The name of the S3 bucket.

        Returns
        -------
        None
        """
        client.put_object(Bucket=bucket, Key=key, Body=fileobj.getvalue())

    ##############################
    # Helper methods
    ##############################

    def _get_bucket(self, root: str) -> str:
        """
        Get the name of the S3 bucket from the URI.

        Returns
        -------
        str
            The name of the S3 bucket.
        """
        return get_bucket_name(root)

    def _get_client(self) -> S3Client:
        """
        Get an S3 client object.

        Returns
        -------
        S3Client
            Returns a client object that interacts with the S3 storage service.
        """
        cfg = self._configurator.get_boto3_client_config()
        return boto3.client("s3", **cfg)

    def _check_factory(self, root: str) -> tuple[S3Client, str]:
        """
        Check if the S3 bucket is accessible by sending a head_bucket request.

        Returns
        -------
        tuple[S3Client, str]
            A tuple containing the S3 client object and the name of the S3 bucket.
        """
        client = self._get_client()
        bucket = self._get_bucket(root)
        self._check_access_to_storage(client, bucket)
        return client, bucket

    def _check_access_to_storage(self, client: S3Client, bucket: str) -> None:
        """
        Check if the S3 bucket is accessible by sending a head_bucket request.

        Parameters
        ----------
        client : S3Client
            The S3 client object.
        bucket : str
            The name of the S3 bucket.

        Returns
        -------
        None

        Raises
        ------
        ClientError:
            If access to the specified bucket is not available.
        """
        try:
            client.head_bucket(Bucket=bucket)
        except ClientError as e:
            raise ClientError("No access to s3 bucket!") from e

    @staticmethod
    def _get_key(path: str) -> str:
        """
        Build key.

        Parameters
        ----------
        path : str
            The source path to get the key from.

        Returns
        -------
        str
            The key.
        """
        key = urlparse(path).path.replace("\\", "/")
        if key.startswith("/"):
            key = key[1:]
        return key

    def _build_key_from_root(self, root: str, paths: list[str]) -> list[str]:
        """
        Method to build object path.

        Parameters
        ----------
        root : str
            The root of the object path.
        paths : list[str]
            The path to build.

        Returns
        -------
        list[str]
            List of keys.
        """
        keys = []
        for path in paths:
            clean_path = self._get_key(path)
            key = self._get_key(f"{root}{clean_path}")
            keys.append(key)
        return keys

    def _list_objects(self, client: S3Client, bucket: str, partition: str) -> list[str]:
        """
        List objects in a S3 partition.

        Parameters
        ----------
        client : S3Client
            The S3 client object.
        bucket : str
            The name of the S3 bucket.
        partition : str
            The partition.

        Returns
        -------
        list[str]
            The list of keys under the partition.
        """
        key = self._get_key(partition)
        file_list = client.list_objects_v2(Bucket=bucket, Prefix=key).get("Contents", [])
        return [f["Key"] for f in file_list]

    @staticmethod
    def is_partition(path: str) -> bool:
        """
        Check if path is a directory or a partition.

        Parameters
        ----------
        path : str
            The path to check.

        Returns
        -------
        bool
        """
        if path.endswith("/"):
            return True
        return False



from __future__ import annotations

import typing

from digitalhub.stores.builder import store_builder

if typing.TYPE_CHECKING:
    from digitalhub.stores._base.store import Store


def get_store(project: str, uri: str, config: dict | None = None) -> Store:
    """
    Get store instance by URI.

    Parameters
    ---------
    project : str
        Project name.
    uri : str
        URI to parse.
    config : dict
        Store configuration.

    Returns
    -------
    Store
        Store instance.
    """
    return store_builder.get(uri, config)



from __future__ import annotations

from typing import Optional

from pydantic import Field

from digitalhub.entities.model._base.spec import ModelSpec, ModelValidator


class ModelSpecHuggingface(ModelSpec):
    """
    ModelSpecHuggingface specifications.
    """

    def __init__(
        self,
        path: str,
        framework: str | None = None,
        algorithm: str | None = None,
        parameters: dict | None = None,
        base_model: str | None = None,
        model_id: str | None = None,
        model_revision: str | None = None,
    ) -> None:
        super().__init__(path, framework, algorithm, parameters)
        self.base_model = base_model
        self.model_id = model_id
        self.model_revision = model_revision


class ModelValidatorHuggingface(ModelValidator):
    """
    ModelValidatorHuggingface validator.
    """

    base_model: Optional[str] = None
    """Base model."""

    placeholder_model_id: str = Field(default=None, alias="model_id")
    """Huggingface model id. If not specified, the model is loaded from the model path."""

    placeholder_model_revision: str = Field(default=None, alias="model_revision")
    """Huggingface model revision."""



from __future__ import annotations

import typing

from digitalhub.configurator.api import get_current_env
from digitalhub.stores.local.store import LocalStore
from digitalhub.stores.remote.store import RemoteStore
from digitalhub.stores.s3.store import S3Store
from digitalhub.stores.sql.store import SqlStore
from digitalhub.utils.uri_utils import SchemeCategory, map_uri_scheme

if typing.TYPE_CHECKING:
    from digitalhub.stores._base.store import Store


def _get_class_from_type(type: str) -> Store:
    """
    Get a store class from its type.

    Parameters
    ----------
    type : str
        Store type.

    Returns
    -------
    Store
        The store class.
    """
    if type == SchemeCategory.LOCAL.value:
        return LocalStore
    if type == SchemeCategory.S3.value:
        return S3Store
    if type == SchemeCategory.REMOTE.value:
        return RemoteStore
    if type == SchemeCategory.SQL.value:
        return SqlStore
    raise ValueError(f"Unknown store type: {type}")


class StoreBuilder:
    """
    Store builder class.
    """

    def __init__(self) -> None:
        self._instances: dict[str, Store] = {}

    def build(self, store_type: str, config: dict | None = None) -> None:
        """
        Build a store instance and register it.

        Parameters
        ----------
        store_type : str
            Store type.
        config : dict
            Store configuration.

        Returns
        -------
        None
        """
        env = get_current_env()
        if env not in self._instances:
            self._instances[env] = {}
        self._instances[env][store_type] = _get_class_from_type(store_type)(config)

    def get(self, uri: str, config: dict | None = None) -> Store:
        """
        Get a store instance by URI.

        Parameters
        ----------
        uri : str
            URI to parse.
        config : dict
            Store configuration.

        Returns
        -------
        Store
            The store instance.
        """
        env = get_current_env()
        store_type = map_uri_scheme(uri)
        try:
            return self._instances[env][store_type]
        except KeyError:
            self.build(store_type, config)
            return self._instances[env][store_type]


store_builder = StoreBuilder()



from __future__ import annotations

import typing
from pathlib import Path

if typing.TYPE_CHECKING:
    from digitalhub.entities.project._base.entity import Project


class Context:
    """
    Context class built forom a `Project` instance. It contains
    some information about the project, such as the project name,
    a client instance (local or non-local), the local context
    project path and information about client locality.
    """

    def __init__(self, project: Project) -> None:
        self.name = project.name
        self.client = project._client
        self.local = project._client.is_local()
        self.root = Path(project.spec.context)
        self.root.mkdir(parents=True, exist_ok=True)

        self.is_running: bool = False
        self._run_ctx: str = None

    def set_run(self, run_ctx: str) -> None:
        """
        Set run identifier.

        Parameters
        ----------
        run_ctx : str
            Run key.

        Returns
        -------
        None
        """
        self.is_running = True
        self._run_ctx = run_ctx

    def unset_run(self) -> None:
        """
        Unset run identifier.

        Returns
        -------
        None
        """
        self.is_running = False
        self._run_ctx = None

    def get_run_ctx(self) -> str:
        """
        Get run identifier.

        Returns
        -------
        str
            Run key.
        """
        return self._run_ctx



from __future__ import annotations

from digitalhub.entities._commons.enums import EntityKinds
from digitalhub.entities.dataitem._base.builder import DataitemBuilder
from digitalhub.entities.dataitem.table.entity import DataitemTable
from digitalhub.entities.dataitem.table.spec import DataitemSpecTable, DataitemValidatorTable
from digitalhub.entities.dataitem.table.status import DataitemStatusTable


class DataitemTableBuilder(DataitemBuilder):
    """
    DataitemTable builder.
    """

    ENTITY_CLASS = DataitemTable
    ENTITY_SPEC_CLASS = DataitemSpecTable
    ENTITY_SPEC_VALIDATOR = DataitemValidatorTable
    ENTITY_STATUS_CLASS = DataitemStatusTable
    ENTITY_KIND = EntityKinds.DATAITEM_TABLE.value



from __future__ import annotations

import os

from digitalhub.configurator.credentials_store import CredentialsStore
from digitalhub.configurator.ini_module import load_from_config, write_config


class EnvConfigurator:
    """
    Configurator object used to configure clients and store credentials.
    """

    def __init__(self) -> None:
        # Store
        self._creds_store = CredentialsStore()

        # Current credentials set (__default by default)
        self._environment = "__default"

    ##############################
    # Public methods
    ##############################

    def set_current_env(self, creds_set: str) -> None:
        """
        Set the current credentials set.

        Parameters
        ----------
        creds_set : str
            Credentials set name.

        Returns
        -------
        None
        """
        self._environment = creds_set

    def get_current_env(self) -> str:
        """
        Get the current credentials set.

        Returns
        -------
        str
            Credentials set name.
        """
        return self._environment

    ##############################
    # Private methods
    ##############################

    def load_var(self, var_name: str) -> str | None:
        """
        Get env variable from credentials store, env or file (in order).

        Parameters
        ----------
        var_name : str
            Environment variable name.

        Returns
        -------
        str | None
            Environment variable value.
        """
        var = self.get_credential(var_name)
        if var is None:
            var = self.load_from_env(var_name)
        if var is None:
            var = self.load_from_config(var_name)
        return var

    def load_from_env(self, var: str) -> str | None:
        """
        Load variable from env.

        Parameters
        ----------
        var : str
            Environment variable name.

        Returns
        -------
        str | None
            Environment variable value.
        """
        env_var = os.getenv(var)
        if env_var != "":
            return env_var

    def load_from_config(self, var: str) -> str | None:
        """
        Load variable from config file.

        Parameters
        ----------
        var : str
            Environment variable name.

        Returns
        -------
        str | None
            Environment variable value.
        """
        return load_from_config(var)

    def write_env(self, key_to_include: list[str] | None = None) -> None:
        """
        Write the env variables to the .dhcore file.
        It will overwrite any existing env variables.

        Parameters
        ----------
        key_to_include : list[str]
            List of keys to include.

        Returns
        -------
        None
        """
        if key_to_include is None:
            key_to_include = []
        creds = self.get_credential_list(key_to_include)
        write_config(creds, self._environment)

    ##############################
    # Credentials store methods
    ##############################

    def set_credential(self, key: str, value: str) -> None:
        """
        Register a credential value.

        Parameters
        ----------
        key : str
            The key.
        value : str
            The value.

        Returns
        -------
        None
        """
        self._creds_store.set(self._environment, key, value)

    def get_credential(self, key: str) -> dict:
        """
        Get single credential value from key.

        Parameters
        ----------
        key : str
            The key.

        Returns
        -------
        dict
            The credential.
        """
        return self._creds_store.get(self._environment, key)

    def get_all_credentials(self) -> dict:
        """
        Get all the credentials from the current credentials set.

        Returns
        -------
        dict
            The credentials.
        """
        return self._creds_store.get_all(self._environment)

    def get_credential_list(self, keys: list[str]) -> list[str]:
        """
        Get the list of credentials.

        Parameters
        ----------
        keys : list[str]
            The list of keys.

        Returns
        -------
        list[str]
            The list of credentials.
        """
        return {k: v for k, v in self.get_all_credentials().items() if k in keys}


# Define global configurator
configurator = EnvConfigurator()



from digitalhub_runtime_kfp.dsl import pipeline_context


def myhandler(ref):
    with pipeline_context() as pc:
        s1 = pc.step(
            name="step1",
            function="function-dbt",
            action="transform",
            inputs=[{"employees": ref}],
            outputs=[{"output_table": "e60"}],
        )

        s2 = pc.step(
            name="step2",
            function="function-dbt",
            action="transform",
            inputs=[{"employees": s1.outputs["e60"]}],
            outputs=[{"output_table": "employees_pipeline"}],
        )

        return s2.outputs["employees_pipeline"]



from __future__ import annotations

from digitalhub.readers.data._base.builder import ReaderBuilder
from digitalhub.readers.data.pandas.reader import DataframeReaderPandas


class ReaderBuilderPandas(ReaderBuilder):
    """
    Pandas reader builder.
    """

    ENGINE = "pandas"
    DATAFRAME_CLASS = "pandas.core.frame.DataFrame"

    def build(self, **kwargs) -> DataframeReaderPandas:
        """
        Build reader object.

        Parameters
        ----------
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        DataframeReaderPandas
            Pandas reader object.
        """
        return DataframeReaderPandas(**kwargs)



from __future__ import annotations

from digitalhub.entities._base.material.status import MaterialStatus


class ArtifactStatus(MaterialStatus):
    """
    Status class for artifact entities.
    """


class ArtifactStatusArtifact(ArtifactStatus):
    """
    Status class for artifact entities.
    """



from digitalhub.entities.artifact.crud import (
    delete_artifact,
    get_artifact,
    get_artifact_versions,
    import_artifact,
    list_artifacts,
    load_artifact,
    log_artifact,
    new_artifact,
    update_artifact,
)
from digitalhub.entities.dataitem.crud import (
    delete_dataitem,
    get_dataitem,
    get_dataitem_versions,
    import_dataitem,
    list_dataitems,
    load_dataitem,
    log_dataitem,
    new_dataitem,
    update_dataitem,
)
from digitalhub.entities.function.crud import (
    delete_function,
    get_function,
    get_function_versions,
    import_function,
    list_functions,
    load_function,
    new_function,
    update_function,
)
from digitalhub.entities.model.crud import (
    delete_model,
    get_model,
    get_model_versions,
    import_model,
    list_models,
    load_model,
    log_model,
    new_model,
    update_model,
)
from digitalhub.entities.project.crud import (
    delete_project,
    get_or_create_project,
    get_project,
    import_project,
    list_projects,
    load_project,
    new_project,
    search_entity,
    update_project,
)
from digitalhub.entities.run.crud import delete_run, get_run, import_run, list_runs, load_run, new_run, update_run
from digitalhub.entities.secret.crud import (
    delete_secret,
    get_secret,
    get_secret_versions,
    import_secret,
    list_secrets,
    load_secret,
    new_secret,
    update_secret,
)
from digitalhub.entities.task.crud import (
    delete_task,
    get_task,
    import_task,
    list_tasks,
    load_task,
    new_task,
    update_task,
)
from digitalhub.entities.workflow.crud import (
    delete_workflow,
    get_workflow,
    get_workflow_versions,
    import_workflow,
    list_workflows,
    load_workflow,
    new_workflow,
    update_workflow,
)

try:
    from digitalhub.entities.model.mlflow.utils import from_mlflow_run, get_mlflow_model_metrics
except ImportError:
    ...

from digitalhub.client.dhcore.utils import refresh_token, set_dhcore_env
from digitalhub.configurator.api import get_current_env, set_current_env

# Register entities into registry
from digitalhub.factory.utils import register_entities, register_runtimes_entities

register_entities()
register_runtimes_entities()



from __future__ import annotations

from enum import Enum
from typing import Optional

from pydantic import BaseModel, ConfigDict, Field


class VolumeType(Enum):
    """
    Volume type.
    """

    PERSISTENT_VOLUME_CLAIM = "persistent_volume_claim"
    EMPTY_DIR = "empty_dir"


class Volume(BaseModel):
    """
    Volume model.
    """

    model_config = ConfigDict(use_enum_values=True)

    volume_type: VolumeType
    """Volume type."""

    name: str
    """Volume name."""

    mount_path: str
    """Volume mount path inside the container."""

    spec: Optional[dict[str, str]] = None
    """Volume spec."""


class NodeSelector(BaseModel):
    """
    NodeSelector model.
    """

    key: str
    """Node selector key."""

    value: str
    """Node selector value."""


class ResourceItem(BaseModel):
    """
    Resource item model.
    """

    requests: str = Field(default=None, pattern=r"[\d]+|^([0-9])+([a-zA-Z])+$")
    """Resource requests."""

    limits: str = Field(default=None, pattern=r"[\d]+|^([0-9])+([a-zA-Z])+$")
    """Resource limits."""


class Resource(BaseModel):
    """
    Resource model.
    """

    cpu: Optional[ResourceItem] = None
    """CPU resource model."""

    mem: Optional[ResourceItem] = None
    """Memory resource model."""

    gpu: Optional[ResourceItem] = None
    """GPU resource model."""


class Env(BaseModel):
    """
    Env variable model.
    """

    name: str
    """Env variable name."""

    value: str
    """Env variable value."""


class Toleration(BaseModel):
    """
    Toleration model.
    """

    key: Optional[str] = None
    """Toleration key."""

    operator: Optional[str] = None
    """Toleration operator."""

    value: Optional[str] = None
    """Toleration value."""

    effect: Optional[str] = None
    """Toleration effect."""

    toleration_seconds: Optional[int] = None
    """Toleration seconds."""


class V1NodeSelectorRequirement(BaseModel):
    key: str
    operator: str
    values: Optional[list[str]] = None


class V1NodeSelectorTerm(BaseModel):
    match_expressions: Optional[list[V1NodeSelectorRequirement]] = None
    match_fields: Optional[list[V1NodeSelectorRequirement]] = None


class V1NodeSelector(BaseModel):
    node_selector_terms: list[V1NodeSelectorTerm]


class V1PreferredSchedulingTerm(BaseModel):
    preference: V1NodeSelector
    weight: int


class V1LabelSelectorRequirement(BaseModel):
    key: str
    operator: str
    values: Optional[list[str]] = None


class V1LabelSelector(BaseModel):
    match_expressions: Optional[list[V1LabelSelectorRequirement]] = None
    match_labels: Optional[dict[str, str]] = None


class V1PodAffinityTerm(BaseModel):
    label_selector: Optional[V1LabelSelector] = None
    match_label_keys: Optional[list[str]] = None
    mismatch_label_keys: Optional[list[str]] = None
    namespace_selector: Optional[V1LabelSelector] = None
    namespaces: Optional[list[str]] = None
    topology_key: Optional[str] = None


class V1WeightedPodAffinityTerm(BaseModel):
    pod_affinity_term: V1PodAffinityTerm
    weight: int


class V1NodeAffinity(BaseModel):
    preferred_during_scheduling_ignored_during_execution: Optional[list[V1PreferredSchedulingTerm]] = None
    required_during_scheduling_ignored_during_execution: Optional[V1NodeSelector] = None


class V1PodAffinity(BaseModel):
    preferred_during_scheduling_ignored_during_execution: Optional[list[V1WeightedPodAffinityTerm]] = None
    required_during_scheduling_ignored_during_execution: Optional[list[V1PodAffinityTerm]] = None


class V1PodAntiAffinity(BaseModel):
    preferred_during_scheduling_ignored_during_execution: Optional[list[V1WeightedPodAffinityTerm]] = None
    required_during_scheduling_ignored_during_execution: Optional[list[V1PodAffinityTerm]] = None


class Affinity(BaseModel):
    """
    Affinity model.
    """

    node_affinity: Optional[V1NodeAffinity] = None
    """Node affinity."""

    pod_affinity: Optional[V1PodAffinity] = None
    """Pod affinity."""

    pod_anti_affinity: Optional[V1PodAntiAffinity] = None
    """Pod anti affinity."""


class K8s(BaseModel):
    """
    Kubernetes resource model.
    """

    node_selector: Optional[list[NodeSelector]] = None
    """Node selector."""

    volumes: Optional[list[Volume]] = None
    """List of volumes."""

    resources: Optional[Resource] = None
    """Resources restrictions."""

    affinity: Optional[Affinity] = None
    """Affinity."""

    tolerations: Optional[list[Toleration]] = None
    """Tolerations."""

    envs: Optional[list[Env]] = None
    """Env variables."""

    secrets: Optional[list[str]] = None
    """List of secret names."""

    profile: Optional[str] = None
    """Profile template."""

    runtime_class: Optional[str] = None
    """Runtime class name."""

    priority_class: Optional[str] = None
    """Priority class."""


class CorePort(BaseModel):
    """
    Port mapper model.
    """

    port: int
    target_port: int


class CoreServiceType(Enum):
    """
    CoreServiceType enum.
    """

    EXTERNAL_NAME = "ExternalName"
    CLUSTER_IP = "ClusterIP"
    NODE_PORT = "NodePort"
    LOAD_BALANCER = "LoadBalancer"


class CorePullPolicy(Enum):
    """
    CorePullPolicy enum.
    """

    ALWAYS = "Always"
    IF_NOT_PRESENT = "IfNotPresent"
    NEVER = "Never"



from __future__ import annotations

import json
from io import BytesIO
from typing import IO, Any

import numpy as np
import pandas as pd
from pandas.errors import ParserError

from digitalhub.entities.dataitem.table.utils import check_preview_size, finalize_preview, prepare_data, prepare_preview
from digitalhub.readers.data._base.reader import DataframeReader
from digitalhub.utils.enums import FileExtensions
from digitalhub.utils.exceptions import ReaderError
from digitalhub.utils.generic_utils import CustomJsonEncoder


class DataframeReaderPandas(DataframeReader):
    """
    Pandas reader class.
    """

    ##############################
    # Read methods
    ##############################

    def read_df(self, path_or_buffer: str | IO, extension: str, **kwargs) -> pd.DataFrame:
        """
        Read DataFrame from path or buffer.

        Parameters
        ----------
        path_or_buffer : str | IO
            Path or buffer to read DataFrame from.
        extension : str
            Extension of the file.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        pd.DataFrame
            Pandas DataFrame.
        """
        if extension == FileExtensions.CSV.value:
            return pd.read_csv(path_or_buffer, **kwargs)
        if extension == FileExtensions.PARQUET.value:
            return pd.read_parquet(path_or_buffer, **kwargs)
        if extension == FileExtensions.JSON.value:
            return pd.read_json(path_or_buffer, **kwargs)
        if extension in (FileExtensions.EXCEL.value, FileExtensions.EXCEL_OLD.value):
            return pd.read_excel(path_or_buffer, **kwargs)
        if extension in (FileExtensions.TXT.value, FileExtensions.FILE.value):
            try:
                return self.read_df(path_or_buffer, FileExtensions.CSV.value, **kwargs)
            except ParserError:
                raise ReaderError(f"Unable to read from {path_or_buffer}.")
        else:
            raise ReaderError(f"Unsupported extension '{extension}' for reading.")

    def read_table(self, sql: str, engine: Any, **kwargs) -> pd.DataFrame:
        """
        Read table from db.

        Parameters
        ----------
        sql : str
            SQL query.
        engine : Any
            SQL Engine.
        **kwargs
            Keyword arguments.

        Returns
        -------
        pd.DataFrame
            Pandas DataFrame.
        """
        return pd.read_sql(sql=sql, con=engine, **kwargs)

    ##############################
    # Write methods
    ##############################

    def write_df(
        self,
        df: pd.DataFrame,
        dst: str | BytesIO,
        extension: str | None = None,
        **kwargs,
    ) -> None:
        """
        Write DataFrame as parquet.

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe to write.
        dst : str | BytesIO
            The destination of the dataframe.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        None
        """
        if extension == FileExtensions.CSV.value:
            return self.write_csv(df, dst, **kwargs)
        if extension == FileExtensions.PARQUET.value:
            return self.write_parquet(df, dst, **kwargs)
        raise ReaderError(f"Unsupported extension '{extension}' for writing.")

    @staticmethod
    def write_csv(df: pd.DataFrame, dst: str | BytesIO, **kwargs) -> None:
        """
        Write DataFrame as csv.

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe to write.
        dst : str | BytesIO
            The destination of the dataframe.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        None
        """
        df.to_csv(dst, index=False, **kwargs)

    @staticmethod
    def write_parquet(df: pd.DataFrame, dst: str | BytesIO, **kwargs) -> None:
        """
        Write DataFrame as parquet.

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe to write.
        dst : str | BytesIO
            The destination of the dataframe.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        None
        """
        df.to_parquet(dst, index=False, **kwargs)

    @staticmethod
    def write_table(df: pd.DataFrame, table: str, engine: Any, schema: str | None = None, **kwargs) -> None:
        """
        Write DataFrame as table.

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe to write.
        table : str
            The destination table.
        engine : Any
            SQL Engine.
        schema : str
            The destination schema.
        **kwargs : dict
            Keyword arguments.

        Returns
        -------
        None
        """
        df.to_sql(table, engine, schema=schema, index=False, **kwargs)

    ##############################
    # Utils
    ##############################

    @staticmethod
    def concat_dfs(dfs: list[pd.DataFrame]) -> pd.DataFrame:
        """
        Concatenate a list of DataFrames.

        Parameters
        ----------
        dfs : list[pd.DataFrame]
            The DataFrames to concatenate.

        Returns
        -------
        pd.DataFrame
            The concatenated DataFrame.
        """
        return pd.concat(dfs, ignore_index=True)

    @staticmethod
    def get_schema(df: pd.DataFrame) -> Any:
        """
        Get schema.

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe.

        Returns
        -------
        Any
            The schema.
        """
        schema = {"fields": []}

        for column_name, dtype in df.dtypes.items():
            field = {"name": str(column_name), "type": ""}

            if pd.api.types.is_integer_dtype(dtype):
                field["type"] = "integer"
            elif pd.api.types.is_float_dtype(dtype):
                field["type"] = "number"
            elif pd.api.types.is_bool_dtype(dtype):
                field["type"] = "boolean"
            elif pd.api.types.is_string_dtype(dtype):
                field["type"] = "string"
            elif pd.api.types.is_datetime64_any_dtype(dtype):
                field["type"] = "datetime"
            else:
                field["type"] = "any"

            schema["fields"].append(field)

        return schema

    @staticmethod
    def get_preview(df: pd.DataFrame) -> dict:
        """
        Get preview.

        Parameters
        ----------
        df : pd.DataFrame
            The dataframe.

        Returns
        -------
        Any
            The preview.
        """
        columns = [str(col) for col, _ in df.dtypes.items()]
        head = df.head(10).replace({np.nan: None})
        data = head.values.tolist()
        prepared_data = prepare_data(data)
        preview = prepare_preview(columns, prepared_data)
        finalizes = finalize_preview(preview, df.shape[0])
        serialized = _serialize_deserialize_preview(finalizes)
        return check_preview_size(serialized)


class PandasJsonEncoder(CustomJsonEncoder):
    """
    JSON pd.Timestamp to ISO format serializer.
    """

    def default(self, obj: Any) -> Any:
        """
        Pandas datetime to ISO format serializer.

        Parameters
        ----------
        obj : Any
            The object to serialize.

        Returns
        -------
        Any
            The serialized object.
        """
        if isinstance(obj, pd.Timestamp):
            return obj.isoformat()
        return super().default(obj)


def _serialize_deserialize_preview(preview: dict) -> dict:
    """
    Serialize and deserialize preview.

    Parameters
    ----------
    preview : dict
        The preview.

    Returns
    -------
    dict
        The serialized preview.
    """
    return json.loads(json.dumps(preview, cls=PandasJsonEncoder))



from __future__ import annotations

import os
import typing
from warnings import warn

from requests import request

from digitalhub.client.dhcore.enums import AuthType, DhcoreEnvVar
from digitalhub.client.dhcore.models import BasicAuth, OAuth2TokenAuth
from digitalhub.configurator.configurator import configurator
from digitalhub.stores.s3.enums import S3StoreEnv
from digitalhub.stores.sql.enums import SqlStoreEnv
from digitalhub.utils.exceptions import ClientError
from digitalhub.utils.uri_utils import has_remote_scheme

if typing.TYPE_CHECKING:
    from requests import Response


# Default key used to store authentication information
AUTH_KEY = "_auth"

# API levels that are supported
MAX_API_LEVEL = 20
MIN_API_LEVEL = 10
LIB_VERSION = 10


class ClientDHCoreConfigurator:
    """
    Configurator object used to configure the client.
    """

    ##############################
    # Configuration methods
    ##############################

    def configure(self, config: dict | None = None) -> None:
        """
        Configure the client attributes with config (given or from
        environment).
        Regarding authentication parameters, the config parameter
        takes precedence over the env variables, and the token
        over the basic auth. Furthermore, the config parameter is
        validated against the proper pydantic model.

        Parameters
        ----------
        config : dict
            Configuration dictionary.

        Returns
        -------
        None
        """
        if config is None:
            self._get_core_endpoint()
            self._get_auth_vars()
            return

        # Read passed config
        # Validate and save credentials
        if config.get("access_token") is not None:
            config = OAuth2TokenAuth(**config)
            for pair in [
                (AUTH_KEY, AuthType.OAUTH2.value),
                (DhcoreEnvVar.ENDPOINT.value, config.endpoint),
                (DhcoreEnvVar.ISSUER.value, config.issuer),
                (DhcoreEnvVar.ACCESS_TOKEN.value, config.access_token),
                (DhcoreEnvVar.REFRESH_TOKEN.value, config.refresh_token),
                (DhcoreEnvVar.CLIENT_ID.value, config.client_id),
            ]:
                configurator.set_credential(*pair)

        elif config.get("user") is not None and config.get("password") is not None:
            config = BasicAuth(**config)
            for pair in [
                (AUTH_KEY, AuthType.BASIC.value),
                (DhcoreEnvVar.ENDPOINT.value, config.endpoint),
                (DhcoreEnvVar.USER.value, config.user),
                (DhcoreEnvVar.PASSWORD.value, config.password),
            ]:
                configurator.set_credential(*pair)

        else:
            raise ClientError("Invalid credentials format.")

    def check_core_version(self, response: Response) -> None:
        """
        Raise an exception if DHCore API version is not supported.

        Parameters
        ----------
        response : Response
            The response object.

        Returns
        -------
        None
        """
        if "X-Api-Level" in response.headers:
            core_api_level = int(response.headers["X-Api-Level"])
            if not (MIN_API_LEVEL <= core_api_level <= MAX_API_LEVEL):
                raise ClientError("Backend API level not supported.")
            if LIB_VERSION < core_api_level:
                warn("Backend API level is higher than library version. You should consider updating the library.")

    def build_url(self, api: str) -> str:
        """
        Build the url.

        Parameters
        ----------
        api : str
            The api to call.

        Returns
        -------
        str
            The url.
        """
        api = api.removeprefix("/")
        return f"{configurator.get_credential(DhcoreEnvVar.ENDPOINT.value)}/{api}"

    ##############################
    # Private methods
    ##############################

    @staticmethod
    def _sanitize_endpoint(endpoint: str) -> str:
        """
        Sanitize the endpoint.

        Returns
        -------
        None
        """
        if not has_remote_scheme(endpoint):
            raise ClientError("Invalid endpoint scheme. Must start with http:// or https://.")

        endpoint = endpoint.strip()
        return endpoint.removesuffix("/")

    def _get_core_endpoint(self) -> None:
        """
        Get the DHCore endpoint from env.

        Returns
        -------
        None

        Raises
        ------
        Exception
            If the endpoint of DHCore is not set in the env variables.
        """
        endpoint = configurator.load_var(DhcoreEnvVar.ENDPOINT.value)
        if endpoint is None:
            raise ClientError("Endpoint not set as environment variables.")
        endpoint = self._sanitize_endpoint(endpoint)
        configurator.set_credential(DhcoreEnvVar.ENDPOINT.value, endpoint)

    def _get_auth_vars(self) -> None:
        """
        Get authentication parameters from the env.

        Returns
        -------
        None
        """
        # Give priority to access token
        access_token = configurator.load_var(DhcoreEnvVar.ACCESS_TOKEN.value)
        if access_token is not None:
            configurator.set_credential(AUTH_KEY, AuthType.OAUTH2.value)
            configurator.set_credential(DhcoreEnvVar.ACCESS_TOKEN.value, access_token)

        # Fallback to basic
        else:
            user = configurator.load_var(DhcoreEnvVar.USER.value)
            password = configurator.load_var(DhcoreEnvVar.PASSWORD.value)
            if user is not None and password is not None:
                configurator.set_credential(AUTH_KEY, AuthType.BASIC.value)
                configurator.set_credential(DhcoreEnvVar.USER.value, user)
                configurator.set_credential(DhcoreEnvVar.PASSWORD.value, password)

    ##############################
    # Auth methods
    ##############################

    def basic_auth(self) -> bool:
        """
        Get basic auth.

        Returns
        -------
        bool
        """
        auth_type = configurator.get_credential(AUTH_KEY)
        return auth_type == AuthType.BASIC.value

    def oauth2_auth(self) -> bool:
        """
        Get oauth2 auth.

        Returns
        -------
        bool
        """
        auth_type = configurator.get_credential(AUTH_KEY)
        return auth_type == AuthType.OAUTH2.value

    def set_request_auth(self, kwargs: dict) -> dict:
        """
        Get the authentication header.

        Parameters
        ----------
        kwargs : dict
            Keyword arguments to pass to the request.

        Returns
        -------
        dict
            Authentication header.
        """
        creds = configurator.get_all_credentials()
        if AUTH_KEY not in creds:
            return kwargs
        if self.basic_auth():
            user = creds[DhcoreEnvVar.USER.value]
            password = creds[DhcoreEnvVar.PASSWORD.value]
            kwargs["auth"] = (user, password)
        elif self.oauth2_auth():
            if "headers" not in kwargs:
                kwargs["headers"] = {}
            access_token = creds[DhcoreEnvVar.ACCESS_TOKEN.value]
            kwargs["headers"]["Authorization"] = f"Bearer {access_token}"
        return kwargs

    def get_new_access_token(self) -> None:
        """
        Get a new access token.

        Returns
        -------
        None
        """
        # Call issuer and get endpoint for
        # refreshing access token
        url = self._get_refresh_endpoint()

        # Call refresh token endpoint
        # Try token from env
        refresh_token = configurator.load_from_env(DhcoreEnvVar.REFRESH_TOKEN.value)
        response = self._call_refresh_token_endpoint(url, refresh_token)

        # Otherwise try token from file
        if response.status_code in (400, 401, 403):
            refresh_token = configurator.load_from_config(DhcoreEnvVar.REFRESH_TOKEN.value)
            response = self._call_refresh_token_endpoint(url, refresh_token)

        response.raise_for_status()
        dict_response = response.json()

        # Read new access token and refresh token
        configurator.set_credential(DhcoreEnvVar.ACCESS_TOKEN.value, dict_response["access_token"])
        configurator.set_credential(DhcoreEnvVar.REFRESH_TOKEN.value, dict_response["refresh_token"])

        # Set new credential in stores
        if (access_key := dict_response.get("aws_access_key_id")) is not None:
            configurator.set_credential(S3StoreEnv.ACCESS_KEY_ID.value, access_key)
            os.environ[S3StoreEnv.ACCESS_KEY_ID.value] = access_key
        if (secret_key := dict_response.get("aws_secret_access_key")) is not None:
            configurator.set_credential(S3StoreEnv.SECRET_ACCESS_KEY.value, secret_key)
            os.environ[S3StoreEnv.SECRET_ACCESS_KEY.value] = secret_key
        if (db_username := dict_response.get("db_username")) is not None:
            configurator.set_credential(SqlStoreEnv.USERNAME.value, db_username)
            os.environ[SqlStoreEnv.USERNAME.value] = db_username
        if (db_password := dict_response.get("db_password")) is not None:
            configurator.set_credential(SqlStoreEnv.PASSWORD.value, db_password)
            os.environ[SqlStoreEnv.PASSWORD.value] = db_password

        # Propagate new access token to config file
        self._write_env()

    def _write_env(self) -> None:
        """
        Write the env variables to the .dhcore.ini file.
        It will overwrite any existing env variables.

        Returns
        -------
        None
        """
        configurator.write_env(
            [
                DhcoreEnvVar.ACCESS_TOKEN.value,
                DhcoreEnvVar.REFRESH_TOKEN.value,
                S3StoreEnv.ACCESS_KEY_ID.value,
                S3StoreEnv.SECRET_ACCESS_KEY.value,
                SqlStoreEnv.USERNAME.value,
                SqlStoreEnv.PASSWORD.value,
            ]
        )

    def _get_refresh_endpoint(self) -> str:
        """
        Get the refresh endpoint.

        Returns
        -------
        str
            Refresh endpoint.
        """
        # Get issuer endpoint
        endpoint_issuer = configurator.load_var(DhcoreEnvVar.ISSUER.value)
        if endpoint_issuer is not None:
            endpoint_issuer = self._sanitize_endpoint(endpoint_issuer)
            configurator.set_credential(DhcoreEnvVar.ISSUER.value, endpoint_issuer)
        else:
            raise ClientError("Issuer endpoint not set.")

        # Standard issuer endpoint path
        url = endpoint_issuer + "/.well-known/openid-configuration"

        # Call issuer to get refresh endpoint
        r = request("GET", url, timeout=60)
        r.raise_for_status()
        return r.json().get("token_endpoint")

    def _call_refresh_token_endpoint(self, url: str, refresh_token: str) -> Response:
        """
        Call the refresh token endpoint.

        Parameters
        ----------
        url : str
            Refresh token endpoint.
        refresh_token : str
            Refresh token.

        Returns
        -------
        Response
            Response object.
        """
        # Get client id
        client_id = configurator.load_var(DhcoreEnvVar.CLIENT_ID.value)
        if client_id is None:
            raise ClientError("Client id not set.")

        # Send request to get new access token
        payload = {
            "grant_type": "refresh_token",
            "client_id": client_id,
            "refresh_token": refresh_token,
            "scope": "openid credentials offline_access",
        }
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        return request("POST", url, data=payload, headers=headers, timeout=60)



from __future__ import annotations

from digitalhub.configurator.configurator import configurator
from digitalhub.stores.sql.enums import SqlStoreEnv
from digitalhub.stores.sql.models import SqlStoreConfig
from digitalhub.utils.exceptions import StoreError


class SqlStoreConfigurator:
    """
    Configure the store by getting the credentials from user
    provided config or from environment.
    """

    def __init__(self, config: dict | None = None) -> None:
        self.configure(config)

    ##############################
    # Configuration methods
    ##############################

    def configure(self, config: dict | None = None) -> None:
        """
        Configure the store by getting the credentials from user
        provided config or from environment.

        Parameters
        ----------
        config : dict
            Configuration dictionary.

        Returns
        -------
        None
        """
        if config is None:
            self._get_config()
        else:
            config: SqlStoreConfig = SqlStoreConfig(**config)
            for pair in [
                (SqlStoreEnv.USERNAME.value, config.user),
                (SqlStoreEnv.PASSWORD.value, config.password),
                (SqlStoreEnv.HOST.value, config.host),
                (SqlStoreEnv.PORT.value, config.port),
                (SqlStoreEnv.DATABASE.value, config.database),
            ]:
                configurator.set_credential(*pair)

    def get_sql_conn_string(self) -> str:
        """
        Get the connection string from environment variables.

        Returns
        -------
        str
            The connection string.
        """
        creds = configurator.get_all_credentials()
        try:
            user = creds[SqlStoreEnv.USERNAME.value]
            password = creds[SqlStoreEnv.PASSWORD.value]
            host = creds[SqlStoreEnv.HOST.value]
            port = creds[SqlStoreEnv.PORT.value]
            database = creds[SqlStoreEnv.DATABASE.value]
            return f"postgresql://{user}:{password}@{host}:{port}/{database}"
        except KeyError as e:
            raise StoreError(f"Missing credentials for SQL store. {str(e)}")

    def _get_config(self) -> None:
        """
        Get the credentials from environment variables.

        Returns
        -------
        None
        """
        user = configurator.load_var(SqlStoreEnv.USERNAME.value)
        password = configurator.load_var(SqlStoreEnv.PASSWORD.value)
        host = configurator.load_var(SqlStoreEnv.HOST.value)
        port = configurator.load_var(SqlStoreEnv.PORT.value)
        database = configurator.load_var(SqlStoreEnv.DATABASE.value)
        if user is None or password is None or host is None or port is None or database is None:
            raise StoreError("Missing credentials for SQL store.")
        configurator.set_credential(SqlStoreEnv.USERNAME.value, user)
        configurator.set_credential(SqlStoreEnv.PASSWORD.value, password)
        configurator.set_credential(SqlStoreEnv.HOST.value, host)
        configurator.set_credential(SqlStoreEnv.PORT.value, port)
        configurator.set_credential(SqlStoreEnv.DATABASE.value, database)



from __future__ import annotations

from digitalhub.entities._base.entity.spec import Spec, SpecValidator
from digitalhub.entities.task._base.models import K8s


class TaskSpec(Spec):
    """TaskSpec specifications."""


class TaskSpecFunction(TaskSpec):
    """TaskSpecFunction specifications."""

    def __init__(
        self,
        function: str,
        node_selector: list[dict] | None = None,
        volumes: list[dict] | None = None,
        resources: dict | None = None,
        affinity: dict | None = None,
        tolerations: list[dict] | None = None,
        envs: list[dict] | None = None,
        secrets: list[str] | None = None,
        profile: str | None = None,
        runtime_class: str | None = None,
        priority_class: str | None = None,
        **kwargs,
    ) -> None:
        self.function = function
        self.node_selector = node_selector
        self.volumes = volumes
        self.resources = resources
        self.affinity = affinity
        self.tolerations = tolerations
        self.envs = envs
        self.secrets = secrets
        self.profile = profile
        self.runtime_class = runtime_class
        self.priority_class = priority_class


class TaskSpecWorkflow(TaskSpec):
    """TaskSpecWorkflow specifications."""

    def __init__(
        self,
        workflow: str,
        node_selector: list[dict] | None = None,
        volumes: list[dict] | None = None,
        resources: dict | None = None,
        affinity: dict | None = None,
        tolerations: list[dict] | None = None,
        envs: list[dict] | None = None,
        secrets: list[str] | None = None,
        profile: str | None = None,
        runtime_class: str | None = None,
        priority_class: str | None = None,
        **kwargs,
    ) -> None:
        self.workflow = workflow
        self.node_selector = node_selector
        self.volumes = volumes
        self.resources = resources
        self.affinity = affinity
        self.tolerations = tolerations
        self.envs = envs
        self.secrets = secrets
        self.profile = profile
        self.runtime_class = runtime_class
        self.priority_class = priority_class


class TaskValidator(SpecValidator):
    """
    TaskValidator validator.
    """


class TaskValidatorFunction(TaskValidator, K8s):
    """
    TaskValidatorFunction validator.
    """

    function: str


class TaskValidatorWorkflow(TaskValidator, K8s):
    """
    TaskValidatorWorkflow validator.
    """

    workflow: str



from __future__ import annotations

import typing

from digitalhub.entities._commons.enums import State
from digitalhub.utils.exceptions import BuilderError

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.entity.status import Status


def build_status(status_cls: Status, **kwargs) -> Status:
    """
    Build entity status object. This method is used to build entity
    status.

    Parameters
    ----------
    status_cls : Status
        Entity status class.
    **kwargs : dict
        Keyword arguments.

    Returns
    -------
    Status
        Entity status object.
    """
    kwargs = parse_arguments(**kwargs)
    return status_cls(**kwargs)


def parse_arguments(**kwargs) -> dict:
    """
    Parse keyword arguments and add default values if necessary.

    Parameters
    ----------
    **kwargs : dict
        Keyword arguments.

    Returns
    -------
    dict
        Keyword arguments with default values.
    """
    state = kwargs.get("state")
    if state is None:
        kwargs["state"] = State.CREATED.value
    else:
        if kwargs["state"] not in State.__members__:
            raise BuilderError(f"Invalid state: {state}")
    return kwargs



from __future__ import annotations


class BuilderError(Exception):
    """
    Raised when incontered errors on builders.
    """


class StoreError(Exception):
    """
    Raised when incontered errors on stores.
    """


class BackendError(Exception):
    """
    Raised when incontered errors from backend.
    """


class EntityNotExistsError(BackendError):
    """
    Raised when entity not found.
    """


class EntityAlreadyExistsError(BackendError):
    """
    Raised when entity already exists.
    """


class MissingSpecError(BackendError):
    """
    Raised when spec is missing in backend.
    """


class UnauthorizedError(BackendError):
    """
    Raised when unauthorized.
    """


class ForbiddenError(BackendError):
    """
    Raised when forbidden.
    """


class BadRequestError(BackendError):
    """
    Raised when bad request.
    """


class EntityError(Exception):
    """
    Raised when incontered errors on entities.
    """


class ContextError(Exception):
    """
    Raised when context errors.
    """


class ReaderError(Exception):
    """
    Raised when incontered errors on readers.
    """


class ClientError(Exception):
    """
    Raised when incontered errors on clients.
    """



from __future__ import annotations

from pydantic import ValidationError

from digitalhub.entities._base.entity.metadata import Metadata, RelationshipValidator
from digitalhub.utils.exceptions import BuilderError
from digitalhub.utils.generic_utils import get_timestamp


def build_metadata(**kwargs) -> Metadata:
    """
    Build entity metadata object. This method is used to build entity
    metadata.

    Parameters
    ----------
    **kwargs : dict
        Keyword arguments for the constructor.

    Returns
    -------
    Metadata
        Metadata object.
    """
    kwargs = parse_arguments(**kwargs)
    return Metadata(**kwargs)


def parse_arguments(**kwargs) -> dict:
    """
    Parse keyword arguments and add default values if necessary.

    Parameters
    ----------
    **kwargs : dict
        Keyword arguments.

    Returns
    -------
    dict
        A dictionary containing the entity metadata attributes.
    """
    if "created" not in kwargs or kwargs["created"] is None:
        kwargs["created"] = get_timestamp()
    if "updated" not in kwargs or kwargs["updated"] is None:
        kwargs["updated"] = kwargs["created"]
    if "relationships" in kwargs:
        if not isinstance(kwargs["relationships"], list):
            raise BuilderError("Invalid relationships format. Must be a list of maps.")
        for relationship in kwargs["relationships"]:
            try:
                RelationshipValidator(**relationship)
            except ValidationError as e:
                raise BuilderError(f"Malformed relationship: {e}") from e
    return kwargs



from __future__ import annotations

from digitalhub.utils.generic_utils import dump_json


class Base:
    """
    Base class for all entities.
    It implements to_dict abd to_json method to represent
    the object as a dictionary/json and an any_setter method to
    set generic attributes coming from a constructor.
    """

    def to_dict(self) -> dict:
        """
        Return object as dict with all non private keys.

        Returns
        -------
        dict
            A dictionary containing the attributes of the entity instance.
        """
        obj = {}
        for key, value in self.__dict__.items():
            if not key.startswith("_") and value is not None:
                if hasattr(value, "to_dict"):
                    sub_dict = value.to_dict()
                    if sub_dict:
                        obj[key] = sub_dict
                else:
                    obj[key] = value
        return obj

    def to_json(self) -> str:
        """
        Return object as json with all non private keys.

        Returns
        -------
        str
            A json string containing the attributes of the entity instance.
        """
        return dump_json(self.to_dict())

    def _any_setter(self, **kwargs) -> None:
        """
        Set any attribute of the object.

        Parameters
        ----------
        **kwargs : dict
            Keyword arguments to be set as attributes.

        Returns
        -------
        None
        """
        for k, v in kwargs.items():
            if k not in self.__dict__:
                setattr(self, k, v)

    def _get_private_attrs(self) -> dict:
        """
        Return all private attributes of the object.

        Returns
        -------
        dict
            A dictionary containing the private attributes of the entity instance.
        """
        return {k: v for k, v in self.__dict__.items() if k.startswith("_")}

    def __repr__(self) -> str:
        """
        Return string representation of the entity object.

        Returns
        -------
        str
            A string representing the entity instance.
        """
        return str(self.to_dict())



from __future__ import annotations

from digitalhub.entities.dataitem._base.status import DataitemStatus


class DataitemStatusDataitem(DataitemStatus):
    """
    DataitemStatusDataitem status.
    """



from __future__ import annotations

import os
from pathlib import Path
from urllib.parse import urlparse

from boto3 import client as boto3_client

from digitalhub.stores.s3.enums import S3StoreEnv

DEFAULT_BUCKET = "datalake"


def get_bucket_name(path: str) -> str:
    """
    Get bucket name from path.

    Parameters
    ----------
    path : str
        The source path to get the key from.

    Returns
    -------
    str
        The bucket name.
    """
    return urlparse(path).netloc


def get_bucket_and_key(path: str) -> tuple[str, str]:
    """
    Get bucket and key from path.

    Parameters
    ----------
    path : str
        The source path to get the key from.

    Returns
    -------
    tuple[str, str]
        The bucket and key.
    """
    parsed = urlparse(path)
    return parsed.netloc, parsed.path


def get_s3_source(bucket: str, key: str, filename: Path) -> None:
    """
    Get S3 source.

    Parameters
    ----------
    bucket : str
        S3 bucket name.
    key : str
        S3 object key.
    filename : Path
        Path where to save the function source.

    Returns
    -------
    None
    """
    s3 = boto3_client("s3", endpoint_url=os.getenv(S3StoreEnv.ENDPOINT_URL.value))
    s3.download_file(bucket, key, filename)


def get_s3_bucket_from_env() -> str | None:
    """
    Function to get S3 bucket name.

    Returns
    -------
    str
        The S3 bucket name.
    """
    return os.getenv(S3StoreEnv.BUCKET_NAME.value, DEFAULT_BUCKET)



from __future__ import annotations

from enum import Enum


class S3StoreEnv(Enum):
    """
    S3Store environment
    """

    ENDPOINT_URL = "S3_ENDPOINT_URL"
    ACCESS_KEY_ID = "AWS_ACCESS_KEY_ID"
    SECRET_ACCESS_KEY = "AWS_SECRET_ACCESS_KEY"
    SESSION_TOKEN = "AWS_SESSION_TOKEN"
    BUCKET_NAME = "S3_BUCKET"
    REGION = "S3_REGION"
    SIGNATURE_VERSION = "S3_SIGNATURE_VERSION"



from __future__ import annotations

from enum import Enum


class SqlStoreEnv(Enum):
    """
    SqlStore environment
    """

    HOST = "DB_HOST"
    PORT = "DB_PORT"
    USERNAME = "DB_USERNAME"
    PASSWORD = "DB_PASSWORD"
    DATABASE = "DB_DATABASE"
    PG_SCHEMA = "DB_SCHEMA"



from __future__ import annotations

from pathlib import Path

from digitalhub.stores.s3.utils import get_s3_bucket_from_env
from digitalhub.utils.file_utils import eval_zip_type
from digitalhub.utils.uri_utils import S3Schemes, has_local_scheme


def eval_local_source(source: str | list[str]) -> None:
    """
    Evaluate if source is local.

    Parameters
    ----------
    source : str | list[str]
        Source(s).

    Returns
    -------
    None
    """
    if isinstance(source, list):
        if not source:
            raise ValueError("Empty list of sources.")
        source_is_local = all(has_local_scheme(s) for s in source)
        for s in source:
            if Path(s).is_dir():
                raise ValueError(f"Invalid source path: {s}. List of paths must be list of files, not directories.")
    else:
        source_is_local = has_local_scheme(source)

    if not source_is_local:
        raise ValueError("Invalid source path. Source must be a local path.")


def eval_zip_sources(source: str | list[str]) -> str:
    """
    Evaluate zip sources.

    Parameters
    ----------
    source : str | list[str]
        Source(s).

    Returns
    -------
    str
        S3Schemes.
    """
    if isinstance(source, list):
        if len(source) > 1:
            return S3Schemes.S3.value
        path = source[0]
    else:
        if Path(source).is_dir():
            return S3Schemes.S3.value
        path = source

    if not eval_zip_type(path):
        return S3Schemes.S3.value
    return S3Schemes.ZIP_S3.value


def build_log_path_from_source(
    project: str,
    entity_type: str,
    name: str,
    uuid: str,
    source: str | list[str],
) -> str:
    """
    Build log path.

    Parameters
    ----------
    project : str
        Project name.
    entity_type : str
        Entity type.
    name : str
        Object name.
    uuid : str
        Object UUID.
    source : str | list[str]
        Source(s).

    Returns
    -------
    str
        Log path.
    """
    scheme = eval_zip_sources(source)
    path = f"{scheme}://{get_s3_bucket_from_env()}/{project}/{entity_type}/{name}/{uuid}"

    if isinstance(source, list) and len(source) >= 1:
        if len(source) > 1:
            path += "/"
        else:
            path += f"/{Path(source[0]).name}"
    elif Path(source).is_dir():
        path += "/"
    elif Path(source).is_file():
        path += f"/{Path(source).name}"

    return path



from __future__ import annotations

import typing

from digitalhub.context.context import Context
from digitalhub.utils.exceptions import ContextError

if typing.TYPE_CHECKING:
    from digitalhub.entities.project._base.entity import Project


class ContextBuilder:
    """
    ContextBuilder class.
    It implements the builder pattern to create a context instance.
    It allows to use multiple projects as context at the same time
    by adding them to the _instances registry with their name.
    """

    def __init__(self) -> None:
        self._instances: dict[str, Context] = {}

    def build(self, project_object: Project, overwrite: bool = False) -> None:
        """
        Add a project as context.

        Parameters
        ----------
        project_object : Project
            The project to add.
        overwrite : bool
            If True, the project will be overwritten if it already exists.

        Returns
        -------
        None
        """
        if (project_object.name not in self._instances) or overwrite:
            self._instances[project_object.name] = Context(project_object)

    def get(self, project: str) -> Context:
        """
        Get a context from project name if it exists.

        Parameters
        ----------
        project : str
            The project name.

        Returns
        -------
        Context
            The project context.

        Raises
        ------
        ValueError
            If the project is not in the context.
        """
        try:
            return self._instances[project]
        except KeyError:
            raise ContextError(f"Context '{project}' not found. Get or create a project named '{project}'.")

    def remove(self, project: str) -> None:
        """
        Remove a project from the context.

        Parameters
        ----------
        project : str
            The project name.

        Returns
        -------
        None
        """
        self._instances.pop(project, None)


context_builder = ContextBuilder()



from __future__ import annotations

import typing

from digitalhub.entities._commons.enums import EntityTypes
from digitalhub.entities._processors.context import context_processor

if typing.TYPE_CHECKING:
    from digitalhub.entities.workflow._base.entity import Workflow


ENTITY_TYPE = EntityTypes.WORKFLOW.value


def new_workflow(
    project: str,
    name: str,
    kind: str,
    uuid: str | None = None,
    description: str | None = None,
    labels: list[str] | None = None,
    embedded: bool = False,
    **kwargs,
) -> Workflow:
    """
    Create a new object.

    Parameters
    ----------
    project : str
        Project name.
    name : str
        Object name.
    uuid : str
        ID of the object.
    description : str
        Description of the object (human readable).
    labels : list[str]
        List of labels.
    embedded : bool
        Flag to determine if object spec must be embedded in project spec.
    **kwargs : dict
        Spec keyword arguments.

    Returns
    -------
    Workflow
        Object instance.

    Examples
    --------
    >>> obj = new_function(project="my-project",
    >>>                    name="my-workflow",
    >>>                    kind="kfp",
    >>>                    code_src="pipeline.py",
    >>>                    handler="pipeline-handler")
    """
    return context_processor.create_context_entity(
        project=project,
        name=name,
        kind=kind,
        uuid=uuid,
        description=description,
        labels=labels,
        embedded=embedded,
        **kwargs,
    )


def get_workflow(
    identifier: str,
    project: str | None = None,
    entity_id: str | None = None,
    **kwargs,
) -> Workflow:
    """
    Get object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.
    entity_id : str
        Entity ID.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    Workflow
        Object instance.

    Examples
    --------
    Using entity key:
    >>> obj = get_workflow("store://my-workflow-key")

    Using entity name:
    >>> obj = get_workflow("my-workflow-name"
    >>>                    project="my-project",
    >>>                    entity_id="my-workflow-id")
    """
    return context_processor.read_context_entity(
        identifier,
        entity_type=ENTITY_TYPE,
        project=project,
        entity_id=entity_id,
        **kwargs,
    )


def get_workflow_versions(
    identifier: str,
    project: str | None = None,
    **kwargs,
) -> list[Workflow]:
    """
    Get object versions from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    list[Workflow]
        List of object instances.

    Examples
    --------
    Using entity key:
    >>> obj = get_workflow_versions("store://my-workflow-key")

    Using entity name:
    >>> obj = get_workflow_versions("my-workflow-name"
    >>>                             project="my-project")
    """
    return context_processor.read_context_entity_versions(
        identifier,
        entity_type=ENTITY_TYPE,
        project=project,
        **kwargs,
    )


def list_workflows(project: str, **kwargs) -> list[Workflow]:
    """
    List all latest version objects from backend.

    Parameters
    ----------
    project : str
        Project name.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    list[Workflow]
        List of object instances.

    Examples
    --------
    >>> objs = list_workflows(project="my-project")
    """
    return context_processor.list_context_entities(
        project=project,
        entity_type=ENTITY_TYPE,
        **kwargs,
    )


def import_workflow(file: str) -> Workflow:
    """
    Import object from a YAML file and create a new object into the backend.

    Parameters
    ----------
    file : str
        Path to YAML file.

    Returns
    -------
    Workflow
        Object instance.

    Examples
    --------
    >>> obj = import_workflow("my-workflow.yaml")
    """
    return context_processor.import_executable_entity(file)


def load_workflow(file: str) -> Workflow:
    """
    Load object from a YAML file and update an existing object into the backend.

    Parameters
    ----------
    file : str
        Path to YAML file.

    Returns
    -------
    Workflow
        Object instance.

    Examples
    --------
    >>> obj = load_workflow("my-workflow.yaml")
    """
    return context_processor.load_executable_entity(file)


def update_workflow(entity: Workflow) -> Workflow:
    """
    Update object. Note that object spec are immutable.

    Parameters
    ----------
    entity : Workflow
        Object to update.

    Returns
    -------
    Workflow
        Entity updated.

    Examples
    --------
    >>> obj = update_workflow(obj)
    """
    return context_processor.update_context_entity(
        project=entity.project,
        entity_type=entity.ENTITY_TYPE,
        entity_id=entity.id,
        entity_dict=entity.to_dict(),
    )


def delete_workflow(
    identifier: str,
    project: str | None = None,
    entity_id: str | None = None,
    delete_all_versions: bool = False,
    cascade: bool = True,
    **kwargs,
) -> dict:
    """
    Delete object from backend.

    Parameters
    ----------
    identifier : str
        Entity key (store://...) or entity name.
    project : str
        Project name.
    entity_id : str
        Entity ID.
    delete_all_versions : bool
        Delete all versions of the named entity. If True, use entity name instead of entity key as identifier.
    cascade : bool
        Cascade delete.
    **kwargs : dict
        Parameters to pass to the API call.

    Returns
    -------
    dict
        Response from backend.

    Examples
    --------
    If delete_all_versions is False:
    >>> obj = delete_workflow("store://my-workflow-key")

    Otherwise:
    >>> obj = delete_workflow("workflow-name",
    >>>                       project="my-project",
    >>>                       delete_all_versions=True)
    """
    return context_processor.delete_context_entity(
        identifier=identifier,
        entity_type=ENTITY_TYPE,
        project=project,
        entity_id=entity_id,
        delete_all_versions=delete_all_versions,
        cascade=cascade,
        **kwargs,
    )



from __future__ import annotations

import typing

from digitalhub.client.builder import client_builder

if typing.TYPE_CHECKING:
    from digitalhub.client._base.client import Client


def get_client(local: bool = False, config: dict | None = None) -> Client:
    """
    Wrapper around ClientBuilder.build.

    Parameters
    ----------
    local : bool
        Whether to create a local client or not.
    config : dict
        DHCore environment configuration.

    Returns
    -------
    Client
        The client instance.
    """
    return client_builder.build(local, config)



from __future__ import annotations

import typing
from typing import Any

from digitalhub.entities._commons.enums import ApiCategories, BackendOperations, Relationship
from digitalhub.entities._processors.utils import (
    get_context_from_identifier,
    get_context_from_project,
    parse_identifier,
)
from digitalhub.factory.api import build_entity_from_dict, build_entity_from_params
from digitalhub.utils.exceptions import EntityAlreadyExistsError, EntityError, EntityNotExistsError
from digitalhub.utils.io_utils import read_yaml
from digitalhub.utils.types import SourcesOrListOfSources

if typing.TYPE_CHECKING:
    from digitalhub.context.context import Context
    from digitalhub.entities._base.context.entity import ContextEntity
    from digitalhub.entities._base.executable.entity import ExecutableEntity
    from digitalhub.entities._base.material.entity import MaterialEntity
    from digitalhub.entities._base.unversioned.entity import UnversionedEntity


class ContextEntityOperationsProcessor:
    """
    Processor for Entity operations.

    This object interacts with the context, check the category of the object,
    and then calls the appropriate method to perform the requested operation.
    Operations can be CRUD, search, list, etc.
    """

    ##############################
    # CRUD context entity
    ##############################

    def _create_context_entity(
        self,
        context: Context,
        entity_type: str,
        entity_dict: dict,
    ) -> dict:
        """
        Create object in backend.

        Parameters
        ----------
        context : Context
            Context instance.
        project : str
            Project name.
        entity_type : str
            Entity type.
        entity_dict : dict
            Object instance.

        Returns
        -------
        dict
            Object instance.
        """
        api = context.client.build_api(
            ApiCategories.CONTEXT.value,
            BackendOperations.CREATE.value,
            project=context.name,
            entity_type=entity_type,
        )
        return context.client.create_object(api, entity_dict)

    def create_context_entity(
        self,
        _entity: ContextEntity | None = None,
        **kwargs,
    ) -> ContextEntity:
        """
        Create object in backend.

        Parameters
        ----------
        **kwargs : dict
            Parameters to pass to entity builder.

        Returns
        -------
        ContextEntity
            Object instance.
        """
        if _entity is not None:
            context = _entity._context()
            obj = _entity
        else:
            context = get_context_from_project(kwargs["project"])
            obj: ContextEntity = build_entity_from_params(**kwargs)
        new_obj = self._create_context_entity(context, obj.ENTITY_TYPE, obj.to_dict())
        return build_entity_from_dict(new_obj)

    def log_material_entity(
        self,
        **kwargs,
    ) -> MaterialEntity:
        """
        Create object in backend and upload file.

        Parameters
        ----------
        **kwargs : dict
            Parameters to pass to entity builder.

        Returns
        -------
        MaterialEntity
            Object instance.
        """
        source: SourcesOrListOfSources = kwargs.pop("source")
        context = get_context_from_project(kwargs["project"])
        obj = build_entity_from_params(**kwargs)
        if context.is_running:
            obj.add_relationship(Relationship.PRODUCEDBY.value, context.get_run_ctx())

        new_obj: MaterialEntity = self._create_context_entity(context, obj.ENTITY_TYPE, obj.to_dict())
        new_obj = build_entity_from_dict(new_obj)
        new_obj.upload(source)
        return new_obj

    def _read_context_entity(
        self,
        context: Context,
        identifier: str,
        entity_type: str | None = None,
        project: str | None = None,
        entity_id: str | None = None,
        **kwargs,
    ) -> dict:
        """
        Read object from backend.

        Parameters
        ----------
        context : Context
            Context instance.
        identifier : str
            Entity key (store://...) or entity name.
        entity_type : str
            Entity type.
        project : str
            Project name.
        entity_id : str
            Entity ID.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        dict
            Object instance.
        """
        project, entity_type, _, entity_name, entity_id = parse_identifier(
            identifier,
            project=project,
            entity_type=entity_type,
            entity_id=entity_id,
        )

        if entity_id is None:
            kwargs["entity_name"] = entity_name
        kwargs = context.client.build_parameters(
            ApiCategories.CONTEXT.value,
            BackendOperations.READ.value,
            **kwargs,
        )

        if entity_id is None:
            api = context.client.build_api(
                ApiCategories.CONTEXT.value,
                BackendOperations.LIST.value,
                project=context.name,
                entity_type=entity_type,
            )
            return context.client.list_first_object(api, **kwargs)

        api = context.client.build_api(
            ApiCategories.CONTEXT.value,
            BackendOperations.READ.value,
            project=context.name,
            entity_type=entity_type,
            entity_id=entity_id,
        )
        return context.client.read_object(api, **kwargs)

    def read_context_entity(
        self,
        identifier: str,
        entity_type: str | None = None,
        project: str | None = None,
        entity_id: str | None = None,
        **kwargs,
    ) -> ContextEntity:
        """
        Read object from backend.

        Parameters
        ----------
        identifier : str
            Entity key (store://...) or entity name.
        entity_type : str
            Entity type.
        project : str
            Project name.
        entity_id : str
            Entity ID.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        VersionedEntity
            Object instance.
        """
        context = get_context_from_identifier(identifier, project)
        obj = self._read_context_entity(
            context,
            identifier,
            entity_type=entity_type,
            project=project,
            entity_id=entity_id,
            **kwargs,
        )
        entity = build_entity_from_dict(obj)
        return self._post_process_get(entity)

    def read_unversioned_entity(
        self,
        identifier: str,
        entity_type: str | None = None,
        project: str | None = None,
        entity_id: str | None = None,
        **kwargs,
    ) -> UnversionedEntity:
        """
        Read object from backend.

        Parameters
        ----------
        identifier : str
            Entity key (store://...) or entity name.
        entity_type : str
            Entity type.
        project : str
            Project name.
        entity_id : str
            Entity ID.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        UnversionedEntity
            Object instance.
        """
        if not identifier.startswith("store://"):
            entity_id = identifier
        else:
            splt = identifier.split(":")
            if len(splt) == 3:
                identifier = f"{splt[0]}:{splt[1]}"
        return self.read_context_entity(
            identifier,
            entity_type=entity_type,
            project=project,
            entity_id=entity_id,
            **kwargs,
        )

    def import_context_entity(
        self,
        file: str,
    ) -> ContextEntity:
        """
        Import object from a YAML file and create a new object into the backend.

        Parameters
        ----------
        file : str
            Path to YAML file.

        Returns
        -------
        ContextEntity
            Object instance.
        """
        dict_obj: dict = read_yaml(file)
        dict_obj["status"] = {}
        context = get_context_from_project(dict_obj["project"])
        obj = build_entity_from_dict(dict_obj)
        try:
            self._create_context_entity(context, obj.ENTITY_TYPE, obj.to_dict())
        except EntityAlreadyExistsError:
            raise EntityError(f"Entity {obj.name} already exists. If you want to update it, use load instead.")
        return obj

    def import_executable_entity(
        self,
        file: str,
    ) -> ExecutableEntity:
        """
        Import object from a YAML file and create a new object into the backend.

        Parameters
        ----------
        file : str
            Path to YAML file.

        Returns
        -------
        ExecutableEntity
            Object instance.
        """
        dict_obj: dict | list[dict] = read_yaml(file)
        if isinstance(dict_obj, list):
            exec_dict = dict_obj[0]
            exec_dict["status"] = {}
            tsk_dicts = []
            for i in dict_obj[1:]:
                i["status"] = {}
                tsk_dicts.append(i)
        else:
            exec_dict = dict_obj
            tsk_dicts = []

        context = get_context_from_project(exec_dict["project"])
        obj: ExecutableEntity = build_entity_from_dict(exec_dict)
        try:
            self._create_context_entity(context, obj.ENTITY_TYPE, obj.to_dict())
        except EntityAlreadyExistsError:
            raise EntityError(f"Entity {obj.name} already exists. If you want to update it, use load instead.")

        obj.import_tasks(tsk_dicts)

        return obj

    def load_context_entity(
        self,
        file: str,
    ) -> ContextEntity:
        """
        Load object from a YAML file and update an existing object into the backend.

        Parameters
        ----------
        file : str
            Path to YAML file.

        Returns
        -------
        ContextEntity
            Object instance.
        """
        dict_obj: dict = read_yaml(file)
        context = get_context_from_project(dict_obj["project"])
        obj: ContextEntity = build_entity_from_dict(dict_obj)
        try:
            self._update_context_entity(context, obj.ENTITY_TYPE, obj.id, obj.to_dict())
        except EntityNotExistsError:
            self._create_context_entity(context, obj.ENTITY_TYPE, obj.to_dict())
        return obj

    def load_executable_entity(
        self,
        file: str,
    ) -> ExecutableEntity:
        """
        Load object from a YAML file and update an existing object into the backend.

        Parameters
        ----------
        file : str
            Path to YAML file.

        Returns
        -------
        ExecutableEntity
            Object instance.
        """
        dict_obj: dict | list[dict] = read_yaml(file)
        if isinstance(dict_obj, list):
            exec_dict = dict_obj[0]
            tsk_dicts = dict_obj[1:]
        else:
            exec_dict = dict_obj
            tsk_dicts = []

        context = get_context_from_project(exec_dict["project"])
        obj: ExecutableEntity = build_entity_from_dict(exec_dict)

        try:
            self._update_context_entity(context, obj.ENTITY_TYPE, obj.id, obj.to_dict())
        except EntityNotExistsError:
            self._create_context_entity(context, obj.ENTITY_TYPE, obj.to_dict())
        obj.import_tasks(tsk_dicts)
        return obj

    def _read_context_entity_versions(
        self,
        context: Context,
        identifier: str,
        entity_type: str | None = None,
        project: str | None = None,
        **kwargs,
    ) -> list[dict]:
        """
        Get all versions object from backend.

        Parameters
        ----------
        context : Context
            Context instance.
        identifier : str
            Entity key (store://...) or entity name.
        entity_type : str
            Entity type.
        project : str
            Project name.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        list[dict]
            Object instances.
        """
        project, entity_type, _, entity_name, _ = parse_identifier(
            identifier,
            project=project,
            entity_type=entity_type,
        )

        kwargs = context.client.build_parameters(
            ApiCategories.CONTEXT.value,
            BackendOperations.READ_ALL_VERSIONS.value,
            entity_name=entity_name,
            **kwargs,
        )

        api = context.client.build_api(
            ApiCategories.CONTEXT.value,
            BackendOperations.LIST.value,
            project=context.name,
            entity_type=entity_type,
        )
        return context.client.list_objects(api, **kwargs)

    def read_context_entity_versions(
        self,
        identifier: str,
        entity_type: str | None = None,
        project: str | None = None,
        **kwargs,
    ) -> list[ContextEntity]:
        """
        Read object versions from backend.

        Parameters
        ----------
        identifier : str
            Entity key (store://...) or entity name.
        entity_type : str
            Entity type.
        project : str
            Project name.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        list[ContextEntity]
            List of object instances.
        """
        context = get_context_from_identifier(identifier, project)
        objs = self._read_context_entity_versions(
            context,
            identifier,
            entity_type=entity_type,
            project=project,
            **kwargs,
        )
        objects = []
        for o in objs:
            entity: ContextEntity = build_entity_from_dict(o)
            entity = self._post_process_get(entity)
            objects.append(entity)
        return objects

    def _list_context_entities(
        self,
        context: Context,
        entity_type: str,
        **kwargs,
    ) -> list[dict]:
        """
        List objects from backend.

        Parameters
        ----------
        context : Context
            Context instance.
        entity_type : str
            Entity type.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        list[dict]
            List of objects.
        """
        api = context.client.build_api(
            ApiCategories.CONTEXT.value,
            BackendOperations.LIST.value,
            project=context.name,
            entity_type=entity_type,
        )
        return context.client.list_objects(api, **kwargs)

    def list_context_entities(
        self,
        project: str,
        entity_type: str,
        **kwargs,
    ) -> list[ContextEntity]:
        """
        List all latest version objects from backend.

        Parameters
        ----------
        project : str
            Project name.
        entity_type : str
            Entity type.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        list[ContextEntity]
            List of object instances.
        """
        context = get_context_from_project(project)
        objs = self._list_context_entities(context, entity_type, **kwargs)
        objects = []
        for o in objs:
            entity: ContextEntity = build_entity_from_dict(o)
            entity = self._post_process_get(entity)
            objects.append(entity)
        return objects

    def _update_context_entity(
        self,
        context: Context,
        entity_type: str,
        entity_id: str,
        entity_dict: dict,
        **kwargs,
    ) -> dict:
        """
        Update object. Note that object spec are immutable.

        Parameters
        ----------
        context : Context
            Context instance.
        entity_type : str
            Entity type.
        entity_id : str
            Entity ID.
        entity_dict : dict
            Entity dictionary.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        dict
            Response from backend.
        """
        api = context.client.build_api(
            ApiCategories.CONTEXT.value,
            BackendOperations.UPDATE.value,
            project=context.name,
            entity_type=entity_type,
            entity_id=entity_id,
        )
        return context.client.update_object(api, entity_dict, **kwargs)

    def update_context_entity(
        self,
        project: str,
        entity_type: str,
        entity_id: str,
        entity_dict: dict,
        **kwargs,
    ) -> ContextEntity:
        """
        Update object. Note that object spec are immutable.

        Parameters
        ----------
        project : str
            Project name.
        entity_type : str
            Entity type.
        entity_id : str
            Entity ID.
        entity_dict : dict
            Entity dictionary.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        ContextEntity
            Object instance.
        """
        context = get_context_from_project(project)
        obj = self._update_context_entity(
            context,
            entity_type,
            entity_id,
            entity_dict,
            **kwargs,
        )
        return build_entity_from_dict(obj)

    def _delete_context_entity(
        self,
        context: Context,
        identifier: str,
        entity_type: str | None = None,
        project: str | None = None,
        entity_id: str | None = None,
        **kwargs,
    ) -> dict:
        """
        Delete object from backend.

        Parameters
        ----------
        context : Context
            Context instance.
        identifier : str
            Entity key (store://...) or entity name.
        entity_type : str
            Entity type.
        project : str
            Project name.
        entity_id : str
            Entity ID.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        dict
            Response from backend.
        """
        project, entity_type, _, entity_name, entity_id = parse_identifier(
            identifier,
            project=project,
            entity_type=entity_type,
            entity_id=entity_id,
        )

        delete_all_versions: bool = kwargs.pop("delete_all_versions", False)
        kwargs = context.client.build_parameters(
            ApiCategories.CONTEXT.value,
            BackendOperations.DELETE.value,
            entity_id=entity_id,
            entity_name=entity_name,
            cascade=kwargs.pop("cascade", None),
            delete_all_versions=delete_all_versions,
            **kwargs,
        )

        if delete_all_versions:
            api = context.client.build_api(
                ApiCategories.CONTEXT.value,
                BackendOperations.LIST.value,
                project=context.name,
                entity_type=entity_type,
            )
        else:
            api = context.client.build_api(
                ApiCategories.CONTEXT.value,
                BackendOperations.DELETE.value,
                project=context.name,
                entity_type=entity_type,
                entity_id=entity_id,
            )
        return context.client.delete_object(api, **kwargs)

    def delete_context_entity(
        self,
        identifier: str,
        project: str | None = None,
        entity_type: str | None = None,
        entity_id: str | None = None,
        **kwargs,
    ) -> dict:
        """
        Delete object from backend.

        Parameters
        ----------
        identifier : str
            Entity key (store://...) or entity name.
        project : str
            Project name.
        entity_type : str
            Entity type.
        entity_id : str
            Entity ID.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        dict
            Response from backend.
        """
        context = get_context_from_identifier(identifier, project)
        return self._delete_context_entity(
            context,
            identifier,
            entity_type,
            context.name,
            entity_id,
            **kwargs,
        )

    def _post_process_get(self, entity: ContextEntity) -> ContextEntity:
        """
        Post process get (files, metrics).

        Parameters
        ----------
        entity : ContextEntity
            Entity to post process.

        Returns
        -------
        ContextEntity
            Post processed entity.
        """
        if hasattr(entity.status, "metrics"):
            entity._get_metrics()
        if hasattr(entity.status, "files"):
            entity._get_files_info()
        return entity

    ##############################
    # Context entity operations
    ##############################

    def _build_context_entity_key(
        self,
        context: Context,
        entity_type: str,
        entity_kind: str,
        entity_name: str,
        entity_id: str | None = None,
    ) -> str:
        """
        Build object key.

        Parameters
        ----------
        context : Context
            Context instance.
        entity_type : str
            Entity type.
        entity_kind : str
            Entity kind.
        entity_name : str
            Entity name.
        entity_id : str
            Entity ID.

        Returns
        -------
        str
            Object key.
        """
        return context.client.build_key(
            ApiCategories.CONTEXT.value,
            project=context.name,
            entity_type=entity_type,
            entity_kind=entity_kind,
            entity_name=entity_name,
            entity_id=entity_id,
        )

    def build_context_entity_key(
        self,
        project: str,
        entity_type: str,
        entity_kind: str,
        entity_name: str,
        entity_id: str | None = None,
    ) -> str:
        """
        Build object key.

        Parameters
        ----------
        project : str
            Project name.
        entity_type : str
            Entity type.
        entity_kind : str
            Entity kind.
        entity_name : str
            Entity name.
        entity_id : str
            Entity ID.

        Returns
        -------
        str
            Object key.
        """
        context = get_context_from_project(project)
        return self._build_context_entity_key(context, entity_type, entity_kind, entity_name, entity_id)

    def read_secret_data(
        self,
        project: str,
        entity_type: str,
        **kwargs,
    ) -> dict:
        """
        Get data from backend.

        Parameters
        ----------
        project : str
            Project name.
        entity_type : str
            Entity type.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        dict
            Response from backend.
        """
        context = get_context_from_project(project)
        api = context.client.build_api(
            ApiCategories.CONTEXT.value,
            BackendOperations.DATA.value,
            project=context.name,
            entity_type=entity_type,
        )
        return context.client.read_object(api, **kwargs)

    def update_secret_data(
        self,
        project: str,
        entity_type: str,
        data: dict,
        **kwargs,
    ) -> None:
        """
        Set data in backend.

        Parameters
        ----------
        project : str
            Project name.
        entity_type : str
            Entity type.
        data : dict
            Data dictionary.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        None
        """
        context = get_context_from_project(project)
        api = context.client.build_api(
            ApiCategories.CONTEXT.value,
            BackendOperations.DATA.value,
            project=context.name,
            entity_type=entity_type,
        )
        context.client.update_object(api, data, **kwargs)

    def read_run_logs(
        self,
        project: str,
        entity_type: str,
        entity_id: str,
        **kwargs,
    ) -> dict:
        """
        Get logs from backend.

        Parameters
        ----------
        project : str
            Project name.
        entity_type : str
            Entity type.
        entity_id : str
            Entity ID.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        dict
            Response from backend.
        """
        context = get_context_from_project(project)
        api = context.client.build_api(
            ApiCategories.CONTEXT.value,
            BackendOperations.LOGS.value,
            project=context.name,
            entity_type=entity_type,
            entity_id=entity_id,
        )
        return context.client.read_object(api, **kwargs)

    def stop_run(
        self,
        project: str,
        entity_type: str,
        entity_id: str,
        **kwargs,
    ) -> None:
        """
        Stop object in backend.

        Parameters
        ----------
        project : str
            Project name.
        entity_type : str
            Entity type.
        entity_id : str
            Entity ID.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        None
        """
        context = get_context_from_project(project)
        api = context.client.build_api(
            ApiCategories.CONTEXT.value,
            BackendOperations.STOP.value,
            project=context.name,
            entity_type=entity_type,
            entity_id=entity_id,
        )
        context.client.create_object(api, **kwargs)

    def resume_run(
        self,
        project: str,
        entity_type: str,
        entity_id: str,
        **kwargs,
    ) -> None:
        """
        Resume object in backend.

        Parameters
        ----------
        project : str
            Project name.
        entity_type : str
            Entity type.
        entity_id : str
            Entity ID.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        None
        """
        context = get_context_from_project(project)
        api = context.client.build_api(
            ApiCategories.CONTEXT.value,
            BackendOperations.RESUME.value,
            project=context.name,
            entity_type=entity_type,
            entity_id=entity_id,
        )
        context.client.create_object(api, **kwargs)

    def read_files_info(
        self,
        project: str,
        entity_type: str,
        entity_id: str,
        **kwargs,
    ) -> list[dict]:
        """
        Get files info from backend.

        Parameters
        ----------
        project : str
            Project name.
        entity_type : str
            Entity type.
        entity_id : str
            Entity ID.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        list[dict]
            Response from backend.
        """
        context = get_context_from_project(project)
        api = context.client.build_api(
            ApiCategories.CONTEXT.value,
            BackendOperations.FILES.value,
            project=context.name,
            entity_type=entity_type,
            entity_id=entity_id,
        )
        return context.client.read_object(api, **kwargs)

    def update_files_info(
        self,
        project: str,
        entity_type: str,
        entity_id: str,
        entity_list: list[dict],
        **kwargs,
    ) -> None:
        """
        Get files info from backend.

        Parameters
        ----------
        project : str
            Project name.
        entity_type : str
            Entity type.
        entity_id : str
            Entity ID.
        entity_list : list[dict]
            Entity list.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        None
        """
        context = get_context_from_project(project)
        api = context.client.build_api(
            ApiCategories.CONTEXT.value,
            BackendOperations.FILES.value,
            project=context.name,
            entity_type=entity_type,
            entity_id=entity_id,
        )
        return context.client.update_object(api, entity_list, **kwargs)

    def read_metrics(
        self,
        project: str,
        entity_type: str,
        entity_id: str,
        metric_name: str | None = None,
        **kwargs,
    ) -> dict:
        """
        Get metrics from backend.

        Parameters
        ----------
        project : str
            Project name.
        entity_type : str
            Entity type.
        entity_id : str
            Entity ID.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        dict
            Response from backend.
        """
        context = get_context_from_project(project)
        api = context.client.build_api(
            ApiCategories.CONTEXT.value,
            BackendOperations.METRICS.value,
            project=context.name,
            entity_type=entity_type,
            entity_id=entity_id,
            metric_name=metric_name,
        )
        return context.client.read_object(api, **kwargs)

    def update_metric(
        self,
        project: str,
        entity_type: str,
        entity_id: str,
        metric_name: str,
        metric_value: Any,
        **kwargs,
    ) -> None:
        """
        Get single metric from backend.

        Parameters
        ----------
        project : str
            Project name.
        entity_type : str
            Entity type.
        entity_id : str
            Entity ID.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        None
        """
        context = get_context_from_project(project)
        api = context.client.build_api(
            ApiCategories.CONTEXT.value,
            BackendOperations.METRICS.value,
            project=context.name,
            entity_type=entity_type,
            entity_id=entity_id,
            metric_name=metric_name,
        )
        context.client.update_object(api, metric_value, **kwargs)

    def _search(
        self,
        context: Context,
        **kwargs,
    ) -> dict:
        """
        Search in backend.

        Parameters
        ----------
        context : Context
            Context instance.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        dict
            Response from backend.
        """
        kwargs = context.client.build_parameters(
            ApiCategories.CONTEXT.value,
            BackendOperations.SEARCH.value,
            **kwargs,
        )
        api = context.client.build_api(
            ApiCategories.CONTEXT.value,
            BackendOperations.SEARCH.value,
            project=context.name,
        )
        entities_dict = context.client.read_object(api, **kwargs)
        return [self.read_context_entity(entity["key"]) for entity in entities_dict["content"]]

    def search_entity(
        self,
        project: str,
        query: str | None = None,
        entity_types: list[str] | None = None,
        name: str | None = None,
        kind: str | None = None,
        created: str | None = None,
        updated: str | None = None,
        description: str | None = None,
        labels: list[str] | None = None,
        **kwargs,
    ) -> list[ContextEntity]:
        """
        Search objects from backend.

        Parameters
        ----------
        project : str
            Project name.
        query : str
            Search query.
        entity_types : list[str]
            Entity types.
        name : str
            Entity name.
        kind : str
            Entity kind.
        created : str
            Entity creation date.
        updated : str
            Entity update date.
        description : str
            Entity description.
        labels : list[str]
            Entity labels.
        **kwargs : dict
            Parameters to pass to the API call.

        Returns
        -------
        list[ContextEntity]
            List of object instances.
        """
        context = get_context_from_project(project)
        return self._search(
            context,
            query=query,
            entity_types=entity_types,
            name=name,
            kind=kind,
            created=created,
            updated=updated,
            description=description,
            labels=labels,
            **kwargs,
        )


context_processor = ContextEntityOperationsProcessor()



from __future__ import annotations

import typing

from digitalhub.entities._base.material.entity import MaterialEntity
from digitalhub.entities._commons.enums import EntityTypes

if typing.TYPE_CHECKING:
    from digitalhub.entities._base.entity.metadata import Metadata
    from digitalhub.entities.dataitem._base.spec import DataitemSpec
    from digitalhub.entities.dataitem._base.status import DataitemStatus


class Dataitem(MaterialEntity):
    """
    A class representing a dataitem.
    """

    ENTITY_TYPE = EntityTypes.DATAITEM.value

    def __init__(
        self,
        project: str,
        name: str,
        uuid: str,
        kind: str,
        metadata: Metadata,
        spec: DataitemSpec,
        status: DataitemStatus,
        user: str | None = None,
    ) -> None:
        super().__init__(project, name, uuid, kind, metadata, spec, status, user)
        self.spec: DataitemSpec
        self.status: DataitemStatus



from __future__ import annotations

import typing
from abc import abstractmethod

from digitalhub.utils.exceptions import BuilderError

if typing.TYPE_CHECKING:
    from digitalhub.readers.data._base.reader import DataframeReader


class ReaderBuilder:
    ENGINE = None
    DATAFRAME_CLASS = None

    def __init__(self):
        if self.ENGINE is None:
            raise BuilderError("ENGINE must be set.")
        if self.DATAFRAME_CLASS is None:
            raise BuilderError("DATAFRAME_CLASS must be set.")

    @abstractmethod
    def build(self, **kwargs) -> DataframeReader:
        """
        Build reader object.
        """



"""
Unit tests for the entity Models
"""

import digitalhub as dh
from digitalhub.entities.model._base.entity import Model


class TestModelCRUD:
    def create_test_dicts(self):
        names = ["test1", "test2", "test3", "test4"]
        uuids = [
            "d150bcca-bb64-451d-8455-dff862254b95",
            "31acdd2d-0c41-428c-b68b-1b133da9e97b",
            "b4a3dfdc-b917-44c4-9a29-613dcf734244",
            "2618d9c4-cd61-440f-aebb-7e5761709f3b",
        ]
        paths = [
            "./data/my_random_forest_model.pkl",
            "s3://bucket/model.pkl",
            "sql://database/schema/linear_regression_model.joblib",
            "https://url.com/bert_base_uncased.pt",
        ]
        kind = ["model", "model", "model", "model"]

        dicts = []
        for i in range(len(names)):
            dicts.append({"name": names[i], "uuid": uuids[i], "path": paths[i], "kind": kind[i]})

        return dicts

    def test_create_delete(self):
        dicts = self.create_test_dicts()
        p = dh.get_or_create_project("test", local=True)
        # Create and delete models
        for i in dicts:
            d = dh.new_model(p.name, **i)
            dh.delete_model(d.key)
            d = dh.new_model(p.name, **i)
            dh.delete_model(d.name, project=p.name, entity_id=d.id)
            d = p.new_model(**i)
            p.delete_model(d.key)
        assert dh.list_models(p.name) == []
        dh.delete_project("test", local=True, clean_context=True)

    def test_list(self):
        dicts = self.create_test_dicts()
        p = dh.get_or_create_project("test", local=True)

        assert dh.list_models(p.name) == []

        for i in dicts:
            dh.new_model(p.name, **i)

        # List models
        l_obj = dh.list_models(p.name)
        assert isinstance(l_obj, list)
        assert len(l_obj) == 4
        for i in l_obj:
            assert isinstance(i, Model)

        # delete listed objects
        for obj in l_obj:
            dh.delete_model(obj.key)

        assert len(dh.list_models(p.name)) == 0

        dh.delete_project("test", clean_context=True, local=True)

    def test_get(self):
        dicts = self.create_test_dicts()
        p = dh.get_or_create_project("test", local=True)

        for i in dicts:
            o1 = dh.new_model(p.name, **i)
            assert isinstance(o1, Model)

            # Get by name and id
            o2 = dh.get_model(o1.name, project=p.name, entity_id=o1.id)
            assert isinstance(o2, Model)
            assert o1.id == o2.id

            # Get by key
            o3 = dh.get_model(o1.key)
            assert isinstance(o3, Model)
            assert o1.id == o3.id

        # delete listed objects
        l_obj = dh.list_models(p.name)
        for obj in l_obj:
            dh.delete_model(obj.key)

        assert len(dh.list_models(p.name)) == 0

        dh.delete_project("test", clean_context=True, local=True)
